// chan_h323.cxx (ng)
/*
 * H.323 protocol handler
 *
 * Copyright (c) 2002-2003 File-Sure Pte. Ltd.
 *
 * The contents of this file are distributed under the terms of
 * the GNU General Public License
 *
 * Software distributed under this License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Original Code is File-Sure Pte. Ltd.
 * -----------------------------------------------------------------
 * $Log: chan_h323.cxx,v $
 * Revision 1.5  2002/05/18 01:08:36  benjamin
 * This release tested with Zapata XP100 FXO card.
 *
 * Revision 1.4  2002/04/18 18:09:33  benjamin
 * chan_h323.cxx
 * h323_hangup   moved write ASTDN statements above hangup: label
 * chan_phone.c
 * parse_gain_value  commented out calculations
 *
 * Revision 1.3  2002/04/16 09:23:41  benjamin
 * This records another merge with lastest Asterisk CVS.
 * Some new files still missing.
 *
*/

/* chan_h323 - driver to h323 interface */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <ptlib.h>

#include <h323pdu.h>
#include <h323con.h>
#include <h323neg.h>
#include <h323caps.h>
#include <h323rtp.h>
#include <h4501.h>
#include <h4504.h>
#include <h450pdu.h>
#include <channels.h>

#include <gsmcodec.h>
#include <lid.h>
#include <mediafmt.h>

#ifdef HAS_VPB
#include <vpblid.h>
#ifndef __GNUC__
#pragma message("Including VPB support.")
#endif
#endif

#include "chan_h323_codec.h"
#include "chan_h323.h"

#define RAND      1+(int)(1000.0*rand()/(RAND_MAX+1.0))
static int usecnt = 0;
static char defext[64];
static pthread_mutex_t usecnt_lock = PTHREAD_MUTEX_INITIALIZER;

GatewayEndPoint *h323ep = (GatewayEndPoint *)NULL;
static PNXGw *instance; 
const char * GatewayEndPoint::type = "H323";
const char * GatewayEndPoint::tdesc = "ITU H.323 Compliant Gateway";

/* Description of CodeXRefInfo:

     const   char *h323_codec_name;
     unsigned int    PnxCodecId;
     H245_AudioCapability::Choices capabilitySubtype;
     int  isFramedCodec;
     int  framesz;
     int  rxframes;   // actually transmit on h323 
     int  txframes;   // actually receiving no h323 side
     unsigned rxFramesInPacket;
     unsigned txFramesInPacket;
     int  in_min;
     int  in_max;   
     int  out_min;
     int  out_max;   
     int  preference;

     These values should be overwritten by correct ones
     defined in the config file

*/
struct CodecXRefInfo CodecXRInfo[]=
{

  { OPAL_G711_ULAW_64K, AST_FORMAT_ULAW, 
	H245_AudioCapability::e_g711Ulaw64k, 0, 240, 30, 30, 240, 30,
               3,6,8,3,2 },
  { OPAL_G711_ALAW_64K, AST_FORMAT_ALAW, 
        H245_AudioCapability::e_g711Alaw64k, 0, 240, 30, 30, 240, 30,
	      3,6,8,3,3 },
  { OPAL_G728,  AST_FORMAT_G728, 
        H245_AudioCapability::e_g728, 1, 96, 30, 30, 96, 20,
	     3,6,8,3,99 },   
  { OPAL_G729,  AST_FORMAT_G729, 
	H245_AudioCapability::e_g729AnnexA, 1, 24, 10, 10, 24, 6,
	    3,6,8,3,99 },
  { OPAL_G729A, AST_FORMAT_G729A, 
	H245_AudioCapability::e_g729AnnexA, 1, 24, 10, 10, 24, 6,
	   3,6,8,3,99 },
  { OPAL_G729AB, AST_FORMAT_G729AB, 
	H245_AudioCapability::e_g729AnnexAwAnnexB, 1, 24, 10, 10, 24, 6,
	  3,6,8,3,99 },
  { OPAL_G7231,  AST_FORMAT_G723_1, 
	H245_AudioCapability::e_g7231, 1, 24, 5, 60, 8, 3,
	  3,6,8,3,1 },
  { OPAL_GSM0610, AST_FORMAT_GSM, 
	H245_AudioCapability::e_gsmFullRate, 1, 33, 30, 30, 7, 4,
	  3,6,8,3,0 },
  { "LPC-10", AST_FORMAT_LPC10, 
	H245_AudioCapability::e_nonStandard, 1, 7, 44, 44, 1, 1,
	  3,6,8,3,99 }, 
  {  OPAL_PCM16, AST_FORMAT_SLINEAR,
	(H245_AudioCapability::Choices)0, -1, 480, 40, 40, 480, 480,
	  -3,5,-3,5,99 },
  { NULL, NULL, 
	(H245_AudioCapability::Choices)0, 0, 0, 0, 0, 0, 0,
	  -3,5,-3,5,99 }
};


struct CodecXRefInfo *IsValidCodec(const char *mediaFmt)
{
    int  x;

    for (x=0; CodecXRInfo[x].h323_codec_name != NULL; x++)
    {
         if (strncmp(mediaFmt,CodecXRInfo[x].h323_codec_name,
                 strlen(CodecXRInfo[x].h323_codec_name)) == 0)
         {
            return &CodecXRInfo[x];	//.PnxCodecId;
         }
    }
    return NULL;
}

static void pthread_usleep(int t)
{
   struct timeval tv;
   t += 1;
   tv.tv_sec = t / 1000;
   tv.tv_usec = (t % 1000) * 1000;
   select(0, NULL, NULL, NULL, &tv);
}

static void pthread_sleep(int s)
{
   pthread_usleep(s*1000);
}

// ---------- start extern C routines ------------------

void semremove(int sid)
{
   union semun {
       int val;                    /* value for SETVAL */
       struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
       unsigned short int *array;  /* array for GETALL, SETALL */
       struct seminfo *__buf;      /* buffer for IPC_INFO */
   } arg;

   semctl(sid,0,IPC_RMID,arg);
}

int semtran(int key)
{
    int sid;

    if ((sid=semget((key_t)key,1,0666 | IPC_CREAT)) == -1)
       ast_log(LOG_WARNING,"semget\n");
    return(sid);
}

static void semcall(int sid, int op)
{
   struct sembuf   sb;

   sb.sem_num = 0;
   sb.sem_op = op;
   sb.sem_flg = 0;
   if (semop(sid,&sb,1) == -1)
      ast_log(LOG_WARNING,"semop\n");
}

void P(int sid)
{
   semcall(sid,-1);
}

void V(int sid)
{
   semcall(sid,1);
}

int usecount()
{
        int res;

        ast_pthread_mutex_lock(&usecnt_lock);
        res = usecnt;
        ast_pthread_mutex_unlock(&usecnt_lock);
        return res;
}

char *description()
{
        return (char *)GatewayEndPoint::tdesc;
}

char *key()
{
    return ASTERISK_GPL_KEY;
}

int load_module()
{
//  int mode = MODE_IMMEDIATE;

     PProcess::PreInitialise(0, NULL, NULL); 
     instance = new PNXGw(); 

     h323ep = new GatewayEndPoint;
     h323ep->Initialise();

     return 0;
}

int unload_module()
{
       struct h323_pvt  *p;

       /* Hangup all interfaces if they have an owner */
       p = h323ep->pvthd;
       while (p) 
       {
           if (p->owner)
	   {
	       cout << "unload_module: softhangup " << p->context << endl;
               ast_softhangup(p->owner);
	   }

           p = p->next;
       }
       h323ep->pvthd = NULL;

       ast_channel_unregister((char *)GatewayEndPoint::type);
       delete instance;
       return 0;
}

void free_h323_pvt(struct h323_pvt *p)
{
    struct h323_pvt *prv;
    struct h323_pvt *tmp;
    char d[128];

    if (p == (struct h323_pvt *)NULL)
	return;
    if (debugtrace > 4)
    {
       strncpy(d,p->dev,30);
       cout << "free_h323_pvt 1-->" << d << endl;
    }

    for (tmp=prv=h323ep->pvthd; 
      		tmp != (struct h323_pvt *)NULL; 
			  prv=tmp, tmp=tmp->next)
    {
        if (strcmp(tmp->dev, p->dev) == 0)
	     break;
    }

    if (debugtrace > 4)
       cout << "free_h323_pvt 2" << endl;

    if (tmp != (struct h323_pvt *)NULL)
    {
       if (debugtrace > 4)
          cout << "free_h323_pvt 2.1" << endl;

       if (tmp == h323ep->pvthd)
           h323ep->pvthd = tmp->next;
       else
          prv->next = tmp->next;
    }
    else return;

    if (debugtrace > 4)
       cout << "free_h323_pvt 3" << endl;

    tmp->isactive = FALSE;

    V(tmp->writesem);
    pthread_usleep(10);
    close(tmp->infd[1]);
    close(tmp->infd[0]);

    pthread_mutex_unlock(&tmp->readlock);
    pthread_mutex_unlock(&tmp->writelock);
    pthread_mutex_unlock(&tmp->watermark);
    pthread_mutex_unlock(&tmp->pnxlock);

    semremove(tmp->writesem);
    
    pthread_mutex_destroy(&tmp->readlock);
    pthread_mutex_destroy(&tmp->writelock);
    pthread_mutex_destroy(&tmp->watermark);
    pthread_mutex_destroy(&tmp->pnxlock);

    if (debugtrace > 4)
       cout << "free_h323_pvt 6 -->" << d << endl;
    ::free(tmp);
}
   

struct h323_pvt *init_h323_pvt(const char *devname)
{
    struct h323_pvt *tmp = (struct h323_pvt *)::malloc(sizeof(struct h323_pvt));
    struct h323_pvt *prv;
    //int   i;

    if (debugtrace > 4)
       cout << "3) init_h323_pvt " << endl;

    if (tmp)
    {
	tmp->magic[0] = (unsigned char)(H323MAGIC >> 8);
	tmp->magic[1] = (unsigned char)(H323MAGIC << 8);
	tmp->max_waitans_tm = 20;
	if (pipe(tmp->infd) == -1)
	{
             ast_log(LOG_WARNING, "Unable to create pipes");
             ::free(tmp);
             return NULL;
        }

        if ((tmp->writesem=semtran(IPC_PRIVATE)) == -1)
        {
            ast_log(LOG_WARNING, "WriteSem");
            ::free(tmp);
            return NULL;
	}

        tmp->waitsem = (int *)NULL;
        pthread_mutex_init(&tmp->readlock,NULL);
        pthread_mutex_init(&tmp->writelock,NULL);
        pthread_mutex_init(&tmp->watermark,NULL);
        pthread_mutex_init(&tmp->pnxlock,NULL);
	tmp->ibuflen = 0;
	tmp->obuflen = 0;
	tmp->readwait = 0;
	tmp->frames_buffered = 0;
	tmp->in_buffered = 0;
	tmp->rx_framerate = 30;   // 30ms for SLINEAR, but shuld get it externally
	tmp->tx_framerate = 30;   // 30ms for SLINEAR, but shuld get it externally
	tmp->rx_frlen = 480;	// default LINEAR
	tmp->tx_frlen = 480;	// default LINEAR
	tmp->maxreadlen = MAXREADLEN;
	tmp->maxwritelen = MAXWRITELEN;
        tmp->in_framerate = 0;
        tmp->out_framerate = 0;  // 2 rate variables for actual rate measure
        tmp->silencesuppression = 0;
        tmp->owner = NULL;
        tmp->lastoutput = -1;
        tmp->lastinput = -1;
	tmp->connection = (GatewayConnection *)NULL;
	tmp->rx_codec = (CodecXRefInfo *)NULL;
	tmp->tx_codec = (CodecXRefInfo *)NULL;
        tmp->incomingCall = 0;
        memset(tmp->ext, 0, sizeof(tmp->ext));
        memset(tmp->dtmf_str, 0, sizeof(tmp->dtmf_str));
        strncpy(tmp->context, "default", sizeof(tmp->context));
        strncpy(tmp->language, "", sizeof(tmp->language));

	// need to implement 
        //strncpy(tmp->callerid, callerid, sizeof(tmp->callerid));
        strncpy(tmp->dev, devname, sizeof(tmp->dev));
	tmp->isactive = TRUE;
	tmp->type = (char *)GatewayEndPoint::type; 
	tmp->next = (struct h323_pvt *)NULL;

	if (h323ep != (GatewayEndPoint *)NULL)
	{
	   if (h323ep->pvthd == (struct h323_pvt *)NULL)
	   {
	      h323ep->pvthd = tmp;
	   }
	   else
	   {
		for (prv=h323ep->pvthd; 
		      prv->next != (struct h323_pvt *)NULL; 
			prv = prv->next);
	      	prv->next = tmp;
	   }
	}
    }
    return tmp;
}


static struct ast_channel *h323_new(struct h323_pvt *i, int state, 
				     char *context, int add_formats)
{
        struct ast_channel *tmp = (struct ast_channel *)NULL;

        if (debugtrace > 4)
           cout << "chan_h323: h323_new " << context << endl;

        tmp = ast_channel_alloc();
        if (tmp) 
	{
                snprintf(tmp->name, sizeof(tmp->name), "H323/%s", i->dev);
                tmp->type = i->type;
                tmp->fds[0] = i->infd[0];
                /* XXX Switching formats silently causes kernel panics XXX */

		tmp->nativeformats = AST_FORMAT_SLINEAR;

		if (add_formats > 0)
		{
		   tmp->nativeformats |= add_formats;
		}

                tmp->pvt->rawreadformat = AST_FORMAT_SLINEAR;
                tmp->state = state;
                if (state == AST_STATE_RING)
                        tmp->rings = 1;
                tmp->pvt->pvt = i;
                tmp->pvt->send_digit = h323_digit;
                tmp->pvt->call = h323_call;
                tmp->pvt->hangup = h323_hangup;
                tmp->pvt->answer = h323_answer;
                tmp->pvt->read = h323_read;
                tmp->pvt->write = h323_write;
                tmp->pvt->exception = h323_exception;

                strncpy(tmp->context, context, sizeof(tmp->context));
                if (strlen(i->ext))
                        strncpy(tmp->exten, i->ext, sizeof(tmp->exten));
                if (strlen(i->language))
                        strncpy(tmp->language, i->language, 
                                                 sizeof(tmp->language));
                if (strlen(i->callerid))
                        tmp->callerid = strdup(i->callerid);
                i->owner = tmp;
		usecnt++;
                ast_update_use_count();

        } else
                ast_log(LOG_WARNING, "Unable to allocate channel structure\n");

        if (debugtrace > 4)
           cout << "h323_new done" << endl;

        return tmp;
}

static struct ast_channel *h323_request(char *type, int format, void *data)
{
        struct h323_pvt *p = NULL;
        struct ast_channel *tmp = (struct ast_channel *)NULL;
        char *name = (char *)data;
	char *s;
//	char *host;

        if (debugtrace > 4)
	    ast_verbose("h323_request: %s=%s format=%d\n",type,
		  (char *)data,format);

        if (name != (char *)NULL && *name != '\0')
	{
	   char n[128];

	   snprintf(n,127,"%s:%04d",name,h323ep->NewSessId());
	   
	   // 1. make a new h323_pvt

	   p = init_h323_pvt((const char *)n);   // context is now unused

	   if (p == NULL)
		return FALSE;

           tmp = h323_new(p,AST_STATE_DOWN,"default",format);

	   // 2. return the new ast_channel struct
        }

	return tmp;
}

static int h323_digit(struct ast_channel *ast, char digit)
{
        struct h323_pvt *p;

	if (debugtrace > 4)
	  ast_log(LOG_WARNING,"h323_digit = %c (%s)\n", digit, ast->name);

        switch(digit) {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
                break;
        case '*':
                break;
        case '#':
                break;
        default:
                ast_log(LOG_WARNING, "Unknown digit '%c'\n", digit);
                return -1;
        }

        if (ast != NULL && ast->pvt != NULL)
        {
           p = (struct h323_pvt *)ast->pvt->pvt;
	   GatewayConnection *myConnection = 
		   (GatewayConnection *)p->connection;
	   myConnection->SendUserInputTone(digit,300,
			     (myConnection->tx_chan)->GetNumber(),0);
        }
        return 0;
}

static int h323_hangup(struct ast_channel *ast)
{
        struct h323_pvt *p;
	GatewayConnection *myConnection = NULL; 

        if (debugtrace > 2)
                ast_log(LOG_DEBUG, "h323_hangup(%s) state=%d\n", ast->name,
			ast->state);
	
        if (ast == NULL || ast->pvt == NULL || !ast->pvt->pvt) 
        {
             if (debugtrace > 0)
               ast_log(LOG_WARNING, "h323_hangup: channel not connected\n");
	     return 1;
        }

        if (debugtrace > 3)
	   cout << "h323_hangup " << ast->name << endl; 
       
        p = (struct h323_pvt *)ast->pvt->pvt;

        pthread_mutex_lock(&p->pnxlock);

	if (ast->state == AST_STATE_DOWN) 
            goto hangup;
 
        if (debugtrace > 3)
	   cout << "h323_hangup 2" << endl; 

        if (debugtrace > 3)
	   cout << "h323_hangup 2.1" << endl; 

	myConnection = (GatewayConnection *)p->connection;

        if (debugtrace > 2)
	   cout << "h323_hangup 3" << endl;

	if (myConnection != NULL)
	{
           if (debugtrace > 2)
	      cout << "h323_hangup 3.1 obuflen=" << p->obuflen <<
                   " state=" << ast->state << endl;

	      if (p->obuflen > 0)
              {
		  int x=0;
		  int d;

                  while (x < 5 && p->obuflen > 0)
		  {
		     d = (p->obuflen/p->tx_frlen)*abs(p->tx_framerate);
                     if (debugtrace > 2)
                        cout << "h323_hangup 3.1.1 wait=" << d << endl;
		     if (d <= 0) d = p->tx_framerate; 
		     pthread_usleep(d);
		     x++;
		  }
	      }


              if (ast->state != AST_STATE_DOWN)
              {
                 ast->state = AST_STATE_DOWN;
                 if (debugtrace > 1)
                      cout << "Clearing Call " << 
				myConnection->gwcallToken << endl;
	         myConnection->Lock();
	         h323ep->ClearCallSynchronous(myConnection->gwcallToken,
				H323Connection::EndedByLocalUser,NULL);
	         myConnection->Unlock();
                 if (debugtrace > 1)
                      cout << "Clearing Call 1.1" << endl;
	      }

              if (debugtrace > 1)
	         cout << "h323_hangup 3.2"  << endl;

	      ast->state = AST_STATE_DOWN;

	      if (p->isactive)
	      {
         	 p->isactive = FALSE;
  	         pthread_mutex_lock(&p->writelock);
	         V(p->writesem);
  	         pthread_mutex_unlock(&p->writelock);
	      }

              if (debugtrace > 1)
	         cout << "h323_hangup 3.3" << endl; 

              pthread_sleep(1);
	}

        write(p->infd[1],"ASTDN\n\0",6);
	ioctl(p->infd[1],FDFLUSH);

hangup:

	ast->state = AST_STATE_DOWN;
	usecnt--;
	if (usecnt < 0) usecnt = 0;
        ast_update_use_count();

        if (debugtrace > 0)
	   cout << "h323_hangup 4 done" << endl; 

        pthread_mutex_unlock(&p->pnxlock);
	return 0;
}


static int h323_answer(struct ast_channel *ast)
{ 
        struct h323_pvt *p;
        p = (struct h323_pvt *)ast->pvt->pvt;

        if (debugtrace > 0)
           cout << "h323_answer name=" << ast->name << endl;

	if (p == (h323_pvt *) NULL)
        {
           if (debugtrace > 3)
	      ast_log(LOG_WARNING,"null pvt\n");
	   return -1;
	}

	pthread_mutex_lock(&p->pnxlock);
	
        if (ast->state >= AST_STATE_UP)
	{
		ast_log(LOG_WARNING,"h323_answer: call not ringing state=%d\n",
			ast->state);
	        pthread_mutex_unlock(&p->pnxlock);
		return -1;
	}

	GatewayConnection *myConnection = 
		   (GatewayConnection *)p->connection;

	if (myConnection != NULL)
	{
            if (ast->state < AST_STATE_RINGING && 
				    !myConnection->IsEstablished())
	    {
           //     ast->state = AST_STATE_UP;

		if (debugtrace > 2)
                   cout << "AnswerCallNow" << endl;

	        myConnection->AnsweringCall(H323Connection::AnswerCallNow);
            }

          //  ast->state = AST_STATE_UP;
	    for (int x=0; x < 40 && p->isactive; x++)
	    {
                if (myConnection->rx_chan != NULL || 
			myConnection->tx_chan != NULL)
                          break;
	        if (debugtrace > 3)
                  cout << "h323_answer WAITING" << endl;
                pthread_sleep(2);
	    }


	    if (myConnection->rx_chan != NULL || 
		       myConnection->rx_chan != NULL )
	    {
            		ast->state = AST_STATE_UP;
			pthread_mutex_unlock(&p->pnxlock);
	   		return 0;
	    }

	}  // if (myConnection ....

	if (debugtrace > 0)
               cout << "h323_answer failed" << endl;
 
	pthread_mutex_unlock(&p->pnxlock);
	return -1;
}

static int h323_call(struct ast_channel *ast, char *dest, int timeout)
{
        struct h323_pvt *p;
	PString token;
        char  *s;

        p = (struct h323_pvt *)ast->pvt->pvt;
  
        if (debugtrace > 2)
	   ast_log(LOG_WARNING,"h323_call: %s\n",dest);

        if ((ast->state != AST_STATE_DOWN) && 
		(ast->state != AST_STATE_RESERVED)) 
	{
                ast_log(LOG_WARNING, "h323_call called on %s,\
neither down nor reserved\n", ast->name);
                return -1;
        }

        if (debugtrace > 4)
          ast_log(LOG_DEBUG, "Connecting %s on %s (%d)\n", dest, ast->name, ast->fds[0]);

	GatewayConnection *myConnection = 
	   (GatewayConnection *)h323ep->MakeCall(dest,token);

        if (myConnection == NULL)
	  return -1;

        if (debugtrace > 2)
          cout << "MakeCall " << token << endl;

	myConnection->gwcallToken = token;
	myConnection->ast_chan = ast; 
        myConnection->h323pvt = p;
        myConnection->line = NULL;
        ast->fds[0] = p->infd[0];
        ast->state = AST_STATE_RINGING;
	p->connection = myConnection;
        
        if (debugtrace > 2)
	  ast_log(LOG_WARNING,"h323_call: %s done\n",dest);
        return 0;
}

static struct ast_frame *h323_exception(struct ast_channel *ast)
{
   struct h323_pvt  *p;
 
   p = (struct h323_pvt *)ast->pvt->pvt;
   p->fr.datalen = 0;
   p->fr.timelen = 0;
   p->fr.data = NULL;
   p->fr.src = p->type;
   p->fr.offset = 0;
   p->fr.mallocd = 0;

   if (debugtrace > 2)
     ast_log(LOG_WARNING,"h323_exception\n");

   switch (p->vexception)
   {
	case H323_OFFHOOK:
		if (ast->state == AST_STATE_RINGING)
		{
			/* They've picked up the phone */
			p->fr.frametype = AST_FRAME_CONTROL;
			p->fr.subclass = AST_CONTROL_ANSWER;
			ast->state = AST_STATE_UP;
                        ast_log(LOG_WARNING,"h323 remote answered\n");
			return &p->fr;
		}  
		else 
		   ast_log(LOG_WARNING, 
                             "Got off hook in weird state %d\n", ast->state);
  		break;

	case H323_DTMF:
		if (option_debug)
			ast_log(LOG_DEBUG, "h323_exception(): DTMF\n");
		break;
    }

    p->fr.frametype = AST_FRAME_NULL;
    p->fr.subclass = 0;
    return &p->fr;
}

//static int dfd = -1;

static struct ast_frame  *h323_read(struct ast_channel *ast)
{
//    void *s,*d;
    int  x,len,t;
    char tmp[13];
    struct h323_pvt *p = (struct h323_pvt *)ast->pvt->pvt;


    if (p == (struct h323_pvt *)NULL || !p->isactive)
    {
        if (debugtrace > 0)
           cout << "h323_read: pvt is down" << endl;
	return NULL;
    }

    p->fr.datalen = 0;
    p->fr.timelen = 0;
    p->fr.data = NULL;
    p->fr.frametype = AST_FRAME_NULL;
    p->fr.src = p->type;
    p->fr.offset = 0;
    p->fr.mallocd = 0;

    if (p == (struct h323_pvt *)NULL || !p->isactive)
    {
        if (debugtrace > 0)
           cout << "h323_read: pvt is down" << endl;
	return NULL;
    }

    while (1)
    {
       if ((x=read(p->infd[0],tmp,6)) < 0)
       {
           if (debugtrace > 0)
	      cout << "h323_read: pipe" << endl;
	   goto read_exit2;
       }

       if (p == (struct h323_pvt *)NULL || !p->isactive)
       {
           if (debugtrace > 0)
              cout << "h323_read: pvt is down" << endl;
	   goto read_exit2;
       }

       if (x == 6)
	    break;
    }
	// flush out the status pipe
    if (strncmp(tmp,"ASTUP",5) == 0)
    {
       ast->rings = 0;
       ast->state = AST_STATE_UP;
       p->fr.frametype = AST_FRAME_CONTROL;
       p->fr.subclass = AST_CONTROL_ANSWER;
       p->vexception = 0;
       x = 0;
       if (debugtrace > 1)
          ast_log(LOG_WARNING,"H323_read: AST_CONTROL_ANSWER\n");
//       if (p->incomingCall != 1)
//          dfd = open("/tmp/dbg_fl",O_WRONLY|O_TRUNC,0666);
       goto read_exit2;
    }

    pthread_mutex_lock(&p->readlock);

    if (p == (struct h323_pvt *)NULL || !p->isactive)
    { 
        if (debugtrace > 0)
           cout << "h323_read: pvt went down" << endl;
	goto read_exit;
    }

    if (strlen(p->dtmf_str) > 0)
    {
        p->fr.frametype = AST_FRAME_DTMF;
        p->fr.subclass = p->dtmf_str[0];
//	p->fr.datalen = p->dtmf_ontime;
//	p->fr.timelen = p->dtmf_offtime;

        for (x=0;x < (int)strlen(p->dtmf_str);x++)
                  p->dtmf_str[x] = p->dtmf_str[x + 1];
        pthread_mutex_unlock(&p->readlock);
	goto read_exit2;
    }

    if (p->ibuflen <= 0 || p->in_buffered <= 0)
       goto read_exit;

 //   ast_log(LOG_WARNING,"h323_read 2\n");
    len = p->ibuflen;	// take a snap, so that write can move ibuflen
    x = 0;

    if (!p->isactive)
       goto read_exit;

    if (len < p->rx_frlen)
        goto read_exit;

    len = p->rx_frlen;
    t = 0;
    p->fr.data = p->frbuf;

    while (p->in_buffered > 0 && p->ibuflen >= len && t < 4)
    {
//    ast_log(LOG_WARNING,"h323_read 3\n");
	p->in_buffered--;
	if (p->in_buffered < 0)
          p->in_buffered = 0;
	memcpy(p->frbuf+(t*len),p->ibuf+(t*len),len);
    	t++;
    	p->ibuflen -= len;
    	if (p->ibuflen < len)    // discard non full frame
	   p->ibuflen = 0;
    }

 //   ast_log(LOG_WARNING,"h323_read 4\n");
    x = len*t;
    if (p->ibuflen > 0)
      memmove(p->ibuf,p->ibuf+x,p->ibuflen);
   
    if (p->rx_codec == NULL)
    { 
       if (debugtrace > 0)
          cout << "WARNING: rx_codec still NULL on h323_read" << endl;
       p->lastinput = AST_FORMAT_SLINEAR;
    }

 //   ast_log(LOG_WARNING,"h323_read done\n");

    p->fr.frametype = AST_FRAME_VOICE;

/* for debugging only

   if (dfd != -1 && p->incomingCall != 1)
    {
      write(dfd,p->frbuf,len);
      ioctl(dfd,FDFLUSH);
    }
*/


    p->fr.offset = 0;	//AST_FRIENDLY_OFFSET;
read_exit:
    pthread_mutex_unlock(&p->readlock);

    if (p->readwait > 0)
    {
        pthread_usleep(p->readwait);
	p->readwait = 0;
    }
    p->fr.datalen = x;
    p->fr.subclass = p->lastinput;
read_exit2:
    return &p->fr;
}

int h323_write_buf(struct h323_pvt *p, char *buf, int len, int frlen)
{
        int space, nframes;

// debug code	
//	cout << "write_buf " << len << " bytes (frlen=" << 
//			frlen << "bytes)" << endl;


	/* Make sure we have enough buffer space to store the frame */

	if (!p->isactive)
	{
            if (debugtrace > 0)
              cout << "h323_write_buf => SYSTEM NOT ACTIVE!!!" << endl;
	//    pthread_mutex_unlock(&p->writelock);
            return -1;
	}

	if (option_verbose > 4)
           ast_log(LOG_WARNING,"wait for lock 1\n");
        pthread_mutex_lock(&p->watermark); 

	if (option_verbose > 4)
           ast_log(LOG_WARNING,"wait for lock 2\n");
	pthread_mutex_lock(&p->writelock);

	if (option_verbose > 4)
           ast_log(LOG_WARNING,"locks okay\n");
	 
        memcpy(p->obuf + p->obuflen, buf, len);
        p->obuflen += len;
	
        {
            space = p->obuflen;
            nframes = 0;
            while (space > frlen)
	    {
		if (nframes >= p->frames_buffered)
		{
        	    p->frames_buffered++;
                    V(p->writesem);
		}
		nframes++;
		space -= frlen;
	    }
	}

	space = p->maxreadlen - p->obuflen;
	pthread_mutex_unlock(&p->writelock);

	if (option_verbose > 3)
            ast_log(LOG_WARNING,"check: space=%d maxreadlen=%d obuflen=%d\n",
			space, p->maxreadlen, p->obuflen);

	if (space >= p->maxreadlen/2)
        {
	    if (option_verbose > 4)
		 ast_log(LOG_WARNING,"unlock watermark\n");
            pthread_mutex_unlock(&p->watermark); 
        }

	if (option_verbose > 4)
	   ast_log(LOG_WARNING, "h323_write_buf return\n");
        return len;
}
 

int h323_write(struct ast_channel *ast, struct ast_frame *frame)
{
	int 	maxfr, res, expected, sofar;
	char	*pos;
	struct h323_pvt  *p = (struct h323_pvt *)ast->pvt->pvt;

// debugging only
//
//     cout << "h323_write framelen=" << frame->datalen;
//     cout << " chan=" << ast->name << endl;
         
    //if (dfd == -1)
    //{
    //     dfd = 0;
//	 ast_log(LOG_WARNING,"10) H323_write started\n");
 //   }

       /* Write a frame of (presumably voice) data */

        if (frame->frametype != AST_FRAME_VOICE) 
        {
                ast_log(LOG_WARNING, "Don't know what to do with  frame type '%d'\n", frame->frametype);
                ast_frfree(frame);
	        ast_softhangup(ast);
                return -1;
        }

        if ( !(frame->subclass&p->lastoutput) )
	{
               ast_log(LOG_WARNING, 
		      "Cannot handle frames in %d format\n
			expecting %d format\n", frame->subclass,p->lastoutput);
               ast_frfree(frame);
               return -1;
        }

	if (ast->state != AST_STATE_UP)
	{
           if (debugtrace > 2)
	     ast_log(LOG_WARNING,"H323_write: state ! UP \n");

           if (h323_answer(ast) == -1)
	   {
	      ast_log(LOG_WARNING,"STATE not UP and answer failed\n");
		return -1;
	   }
	   ast->state = AST_STATE_UP;
	}

        sofar = 0;
	maxfr = p->tx_frlen;    
        pos = (char *)frame->data;
        while (sofar < frame->datalen && p->isactive) 
	{
                /* Write in no more than maxfr sized frames */
                expected = frame->datalen - sofar;
                if (maxfr < expected)
                        expected = maxfr;

                res = h323_write_buf(p, pos, expected, maxfr);

                if (res != expected) 
		{
                     if (res < 0)
                          ast_log(LOG_WARNING, "Write returned error (%s)\n", strerror(errno));
/*
 * Card is in non-blocking mode now and it works well now, but there are
 * lot of messages like this. So, this message is temporarily disabled.
 */
#if 0
                     else
                          ast_log(LOG_WARNING, "Only wrote %d of %d bytes\n", res, frame->datalen);
#endif
                     return -1;
                }

               sofar += res;
               pos += res;
        }

// debug only
//
//        cout << "h323_write " << sofar << " bytes " << endl;

        return 0;
}

// ---------------- end extern C --------------------------


// ---------------- start C++ routines --------------------
       

void GatewayEndPoint::OnConnectionEstablished(H323Connection & connection,
					const PString & token)
{
     struct h323_pvt   *p;
     GatewayConnection & myConnection = (GatewayConnection &)connection;
     struct ast_channel  *ast;

     if (debugtrace > 1)
       cout << "9) OnConnectionEstablished "  << endl;

     p = (struct h323_pvt *)myConnection.h323pvt;
     ast = myConnection.ast_chan;

     if (ast != NULL && p != NULL && ast->state != AST_STATE_DOWN)
     {
        ast->rings = 0;

        if (debugtrace > 2)
           cout << "9.1) pvt=" << p->dev << endl;

        p->vexception = H323_OFFHOOK;

        if (myConnection.rx_chan != NULL && myConnection.tx_chan != NULL)
        {
           if (debugtrace > 3)
              cout << "AST_STATE_UP flag set" << endl;

           ast->state = AST_STATE_UP;
           write(p->infd[1],"ASTUP\n\0",6);
           ioctl(p->infd[1],FDFLUSH);
        }
     }
     // pthread_sleep(1);
}


GatewayEndPoint::GatewayEndPoint()
{
  terminalType = e_GatewayOnly;
  pvthd = (struct h323_pvt *)NULL;
  sessid = 1;
}

GatewayEndPoint::~GatewayEndPoint()
{
    struct h323_pvt *p;

    if (debugtrace > 0)
        cout << "destroy GatewayEndPoint" << endl; 

    p = h323ep->pvthd;
    while (p) 
    {
        if (p->owner)
        {
            if (debugtrace > 0)
	       cout << "unload_module: softhangup " << p->context << endl;
            ast_softhangup(p->owner);
	}
	p->connection = NULL;
        p = p->next;
    }
    h323ep->pvthd = NULL;
     
    if (debugtrace > 1)
      cout << "Done ~GatewayEndPoint" << endl; 
}


BOOL GatewayEndPoint::Initialise()
{
   struct ast_config *cfg;
   struct ast_variable *v;
   PIPSocket::Address *interfaceAddress = NULL;
   int findgatekeeper=0;
   PString *gatekeeperAddress = (PString *)NULL;
   PString *gatekeeperId = (PString *)NULL;
   char *s;
   struct CodecXRefInfo *f;

   ast_verbose( "\n\nPNX (H.323 IP-PBX), " PNX_VERSION "\nCopyright (C) 2001-2003 File-Sure Pte Ltd.\n");
   ast_verbose( "Written by Benjamin Har <benjamin@file-sure.com>\n");
   ast_verbose( "=========================================================================\n");

   cfg = ast_load("h323.conf");
   v = ast_variable_browse(cfg, "parameters");
   jitter = 150;   /*Default jitter buffer length is 150ms   */
   //DisableFastStart(0);

   while(v) 
   {
         /* Create the interface list */
        if (!strcasecmp(v->name, "jitter")) 
	{
	    int newjitter = atoi(v->value);
	    if (newjitter >= 20 && newjitter <=10000)
		jitter = newjitter;
	    else
		ast_log(LOG_WARNING,
		   "Jitter should be >= 20 ms and <= 10s, using %dms",jitter);
	    goto nextv;
	}

	if (!strcasecmp(v->name,"IpTOS"))
	{
            int  i=(int)strtol(v->value,0,16); 
	    
            SetRtpIpTypeofService((unsigned)i);
	    goto nextv;
	}

	if (!strcasecmp(v->name,"context"))
	{
	    strncpy(defext,v->value,sizeof(defext));
	    goto nextv;
	}
	
	if (!strcasecmp(v->name,"company_name"))
	{
	    SetLocalUserName(PString(v->value));
	    goto nextv;
	}

        if (!strcasecmp(v->name,"country_code"))
        {
	   t35CountryCode = atoi(v->value);
	}

        if (!strcasecmp(v->name,"disable_faststart"))
        {
	   disableFastStart = ( !strcasecmp(v->value,"YES") ? TRUE : FALSE );
	}

	if (!strcasecmp(v->name,"ipaddress"))
	{
	    if (strlen(v->value) > 0)
	        interfaceAddress = new PIPSocket::Address(PString(v->value));
	}

	if (!strcasecmp(v->name,"gatekeeperip"))
	{
	    if (strlen(v->value) > 0)
	        gatekeeperAddress = new PString(v->value);
	}

	if (!strcasecmp(v->name,"gatekeeperid"))
	{
	    if (strlen(v->value) > 0)
	        gatekeeperId = new PString(v->value);
	}

        if (!strcasecmp(v->name,"findgatekeeper"))
        {
	   findgatekeeper = ( !strcasecmp(v->value,"YES") ? 1 : 0 );
	}

nextv: 
	v = v->next;

   }  // while (v) ...

   if (option_verbose > 3)
      cout << "Done parsing" << endl; 

   accessControl.LoadHostsAccess();

  H323ListenerTCP * listener;

//  if (interfaceAddress == NULL)
//	interfaceAddress = new PIPSocket::Address();

  if (interfaceAddress == NULL || (interfaceAddress->AsString()).IsEmpty())
  {
    interfaceAddress = new PIPSocket::Address("0.0.0.0");
  }

  listener  = new H323ListenerTCP(*this, *interfaceAddress);
  cout << "H323 Starting Listener on " << interfaceAddress->AsString() << endl;

  if (!StartListener(listener)) {
    cout <<  "Could not open H.323 listener port on "
         << listener->GetListenerPort() << endl;
    ast_destroy(cfg);
    pthread_sleep(5);
    unload_module();
    return FALSE;
  }

  pthread_sleep(3);

  if (option_verbose > 3)
     cout << "ast_channel_register " << endl; 

  if (ast_channel_register((char *)type,(char *)tdesc, AST_FORMAT_SLINEAR,
					h323_request)) 
  {
	ast_log(LOG_ERROR, "Unable to register channel class %s\n", type);
        ast_destroy(cfg);
        unload_module();
        return -1;
  }

  SetCapability(0, 0, new H323_GSM0610Capability);
  SetCapability(0, 0, new H323_G711Capability(H323_G711Capability::muLaw));

  v = ast_variable_browse(cfg, "codecs");
  while (v)
  {
     	if (!strncasecmp(v->value,"ON",2))
	{
	     AddAllCapabilities(0,0,v->name);
	     if (option_verbose > 3)
	        cout << "Added Capability: " << v->name << endl;
	}
	s=strtok(v->value,"|");
	if (s != NULL && (f=IsValidCodec(v->name)) != (CodecXRefInfo *)NULL)
	{
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
	     f->framesz = atoi(s);
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
	     f->rxframes = atoi(s);
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
	     f->txframes = atoi(s);
             if (option_verbose > 3)
	     {
  	        cout << f->h323_codec_name << " rxframes=" << f->rxframes << endl;
  	        cout << f->h323_codec_name << " txframes=" << f->txframes << endl;
	     }
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
	     f->rxFramesInPacket = atoi(s);
             if (option_verbose > 3)
  	         cout << f->h323_codec_name << " rxFramesInPkt=" << f->rxFramesInPacket << endl;
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
	     f->txFramesInPacket = atoi(s);
             if (option_verbose > 3)
  	            cout << f->h323_codec_name << " txFramesInPkt=" << f->txFramesInPacket << endl;

	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
             f->in_min = atoi(s);
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
             f->in_max = atoi(s);
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
             f->out_min = atoi(s);
	     if ((s=strtok(NULL,"|")) == NULL) goto nextcodec;
             f->out_max = atoi(s);

	     if ((s=strtok(NULL,"\n")) != NULL && *s != '\0')
	     {
		f->preference = atoi(s);
	     }
	}
nextcodec:
	v = v->next;
  }
	
  AddAllUserInputCapabilities(0, 1);

  // NOT YET TESTED

  // Gatekeeper registration

  H323TransportUDP * rasChannel;

  rasChannel  = new H323TransportUDP(*this, *interfaceAddress);

  if (gatekeeperAddress != NULL && !gatekeeperAddress->IsEmpty()) 
  {
    if (SetGatekeeper(*gatekeeperAddress, rasChannel))
      cout << "Gatekeeper set: " << *gatekeeper << endl;
    else 
    {
      cerr << "Error registering with gatekeeper at \"" << *gatekeeper << '"' << endl;
      return FALSE;
    }
  } 
  else if (gatekeeperId != NULL && !gatekeeperId->IsEmpty()) 
  {
    cout << "Searching for gatekeeper with id \"" << gatekeeperId << "\" ..." << endl;
    if (LocateGatekeeper(*gatekeeperId, rasChannel))
      cout << "Gatekeeper set: " << *gatekeeper << endl;
    else 
    {
      cerr << "Error registering with gatekeeper at \"" << gatekeeperId << '"' << endl;
      return FALSE;
    }
  } 
  else if (findgatekeeper) 
  {
    cout << "Searching for gatekeeper..." << endl;
    if (DiscoverGatekeeper(rasChannel))
      cout << "Gatekeeper found: " << *gatekeeper << endl;
    else 
    {
      cerr << "No gatekeeper found." << endl;
      return FALSE;
    }
  } 
  else 
  {
    // Since we didn't call any gatekeeper registration function,
    // we have to delete the unused rasChannel ourself.
    delete rasChannel;
  }

  ast_destroy(cfg);
  return 0;
}
   

void GatewayEndPoint::SetEndpointTypeInfo( H225_EndpointType & info ) const
{
  if (debugtrace > 0)
    cout << "SetEndPointTypeInfo" << endl;

  H323EndPoint::SetEndpointTypeInfo(info);
  info.m_gateway.IncludeOptionalField(H225_GatewayInfo::e_protocol);
  info.m_gateway.m_protocol.SetSize(1);
  H225_SupportedProtocols &protocol=info.m_gateway.m_protocol[0];
  protocol.SetTag(H225_SupportedProtocols::e_voice);
  PINDEX as=SupportedPrefix.GetSize();
  ((H225_VoiceCaps &)protocol).m_supportedPrefixes.SetSize(as);
  for (PINDEX p=0; p<as; p++)
    H323SetAliasAddress(SupportedPrefix[p], 
             ((H225_VoiceCaps &)protocol).m_supportedPrefixes[p].m_prefix);
}

//
// Rewrite E164 number according to the dial plan
// Format:
//  prefix_match=strip,prefix[,suffix]
//
//         strip    number of digits to strip
//         prefix   prefix to insert
//         suffix   suffix to append
//
// Example:
//  09=1,886      0953378875 --> 8869378875
//  07=2,,0       07123      --> 1230
//
PString GatewayEndPoint::RewriteE164(const PString &dialedDigits, 
                                        const PStringToString &myplan)
{
  if (option_verbose > 0)
    cout << "RewriteE164-1" << endl;

  for (PINDEX i=dialedDigits.GetLength(); i>0; i--) {
    PString key = dialedDigits.Left(i);
    if (myplan.Contains(key)) {
      PStringArray rule = myplan[key].Tokenise(",");
      PString result = rule[1] + dialedDigits.Mid(rule[0].AsInteger());
      if (!rule[2])
        result += rule[2];

      PTRACE(1, "Prefix " << key << " matched, rewrite " << dialedDigits << " to " << result << endl);
      return result;
    }
  }
  // no match
  return dialedDigits;
}

PString GatewayEndPoint::RewriteE164(const PString &dialedDigits)
{
  if (option_verbose > 3)
       cout << "RewriteE164-2" << endl;

  static PStringToString DialPlan(PConfig("DialPlan").GetAllKeyValues());
 // need to change to read dial plans from h323.conf file
 // 
  return RewriteE164(dialedDigits, DialPlan);
}

H323Connection * GatewayEndPoint::CreateConnection(unsigned callReference)
{
  return new GatewayConnection(*this, callReference);
}


// When incoming call arrives, it is designated for one of
// extension numbers or the main line. We will get this
// from the E164 info field. We strip out what is designated
// as the station id and the remain is the extension
// 

BOOL GatewayEndPoint::OnIncomingCall(H323Connection & connection,
                                    const H323SignalPDU & setupPDU,
                                    H323SignalPDU &)
{
   struct h323_pvt *p;
   H323TransportAddress address = 
		connection.GetControlChannel().GetRemoteAddress();
   PIPSocket::Address ip;
   WORD port;
   BOOL res = TRUE;
   int formats = 0;

   if (address.GetIpAndPort(ip, port) && !accessControl.IsAllowed(ip)) {
     if (debugtrace > 1)
        cout << "Incoming H.323 call from " << ip << " is denied." << endl;
     return FALSE;
   }
   
   if (debugtrace > 0)
      cout << "1) OnIncomingCall from " << ip << endl;

   GatewayConnection & myConnection = (GatewayConnection &)connection;
   
   PString number;

   if (debugtrace > 0)
      cout << "1.1) After number init" << endl;

   if (!setupPDU.GetDestinationE164(number))
   {
       if (debugtrace > 1)
          cout << "2) OnIncomingCall number set to default" << endl;
        number = PString((const char *)defext);    // dial the main line
   }
   
   myConnection.remote_endpoint = address;
   myConnection.gwcallToken = connection.GetCallToken();
   if (debugtrace > 0)
      cout << "2) gwcallToken is " << myConnection.gwcallToken << endl;
   p = myConnection.h323pvt = init_h323_pvt(myConnection.gwcallToken);
   p->connection = &myConnection;

   myConnection.callState = H323Connection::AnswerCallPending;

   if (number)
   {
      if (debugtrace > 1)
         cout << "4) OnIncomingCall Dialing " << number << endl;

      if (number != defext)
      {
         PString dialed(RewriteE164(number)); 

         if (debugtrace > 2)
             cout << "4.1) RewriteE164=" << dialed << endl;
         strcat(p->ext,(const char *)dialed); 
      }
      else
         strcat(p->ext,number);
   } 	// if (!number

   if (ast_exists_extension(NULL,p->context, p->ext,1, p->callerid))
   {
       if (debugtrace > 1)
         cout << "6) creating ast channel for call " << p->ext << endl;
       //myConnection.ast_chan = h323_new(p,AST_STATE_RING,p->context);
   }
   else 
   {
       // There is nothing in the specified extension that can match anymore.  Try the default 
        if (ast_exists_extension(NULL, defext, 
                                 p->ext, 1, p->callerid)) 
	{
            strcpy(p->context,defext);
	    // Check the default, too... 
            if (debugtrace > 2)
	      cout << "7) using default connection" << endl;
	}
	else
	{
  	   myConnection.callState = H323Connection::AnswerCallDenied;
	   return FALSE;
	}
   }

   myConnection.line = new OpalPnXDevice(myConnection.h323pvt);

   p->incomingCall = 1;


   if (debugtrace > 0)
      cout << "5) OnIncomingCall Done" << endl;

   return res;
}


BOOL GatewayConnection::CheckFastStart()
{
   BOOL res = FALSE;

   if (fastStartState != FastStartDisabled) 
   {
     if (debugtrace > 1)
       cout << " FAST START is ON" << endl;
     res = TRUE;
   }
   else
     if (debugtrace > 1)
        cout << "FAST START is OFF" << endl;

   return res;
}


H323Connection::AnswerCallResponse 
  GatewayEndPoint::OnAnswerCall(H323Connection & connection,
                                const PString & /*callerName*/,
                                const H323SignalPDU & /*setupPDU*/,
                                H323SignalPDU & connectPDU )
{
    int  f=(int)NULL;

/*
  if (!convertTones)
    return H323Connection::AnswerCallNow;
*/

   GatewayConnection & myConnection = (GatewayConnection &)connection;

   if (myConnection.ast_chan == NULL)
   {
       myConnection.gwcallToken = connection.GetCallToken();

      if (debugtrace > 0)
         cout << "OnAnswerCall " << myConnection.gwcallToken << " remotecaps=" << 
         connection.GetRemoteCapabilities().GetSize() << endl;
  

   	myConnection.ast_chan = h323_new(myConnection.h323pvt, AST_STATE_RING,
				 myConnection.h323pvt->context,f); 


   	if (ast_pbx_start(myConnection.ast_chan)) 
   	{
          ast_log(LOG_WARNING, 
	        "Unable to start PBX on %s\n", myConnection.ast_chan->name);
          ast_channel_free(myConnection.ast_chan);
          return H323Connection::AnswerCallDenied;
   	}

   	if (debugtrace > 0)
         	cout << "OnAnswerCall Finish= CallAnswered" << endl;

   }

   if (myConnection.ast_chan == NULL ||
		myConnection.ast_chan->state == AST_STATE_DOWN)
        return H323Connection::AnswerCallDenied;

   return H323Connection::AnswerCallPending;
}


BOOL GatewayEndPoint::OpenAudioChannel(H323Connection & connection,
                                      BOOL isEncoding,
                                      unsigned bufferSize,
                                      H323AudioCodec & codec)
{
  OpalLineChannel *lChan = NULL;
  GatewayConnection & myConnection = (GatewayConnection &)connection;

  if (debugtrace > 0)
    cout << "8) OpenAudioChannel codec=" << codec << endl;

  if (myConnection.ast_chan == (struct ast_channel *)NULL )
  {
     if (debugtrace > 1)
        cout << "8.1) No ast channel" << endl;
	return FALSE;
  }

  if (myConnection.h323pvt == (struct h323_pvt *)NULL ||
	myConnection.h323pvt->owner == NULL)
  {
     if (debugtrace > 1)
        cout << "8.2) Invalid pvt struct (" << myConnection.gwcallToken << ")" << endl;
	return FALSE;
  }

  if (myConnection.line == (OpalPnXDevice *)NULL)
  {
     if (debugtrace > 1)
      cout << "Recreated line" << endl;
      myConnection.line = new OpalPnXDevice(myConnection.h323pvt);
  }
   
  codec.SetSilenceDetectionMode(H323AudioCodec::AdaptiveSilenceDetection);
  

  lChan = new OpalLineChannel(myConnection.line->GetDevice(),                                    myConnection.line->GetLineNumber(), codec);


  if (debugtrace > 0)
    cout << "8.3) OpenAudioChannel returns" << endl;

  return codec.AttachChannel(lChan);
}


void GatewayEndPoint::OnConnectionCleared(H323Connection & connection,
                                         const PString &token)
{
  struct ast_channel *chan;
  GatewayConnection & myConnection = (GatewayConnection &)connection;
  struct h323_pvt *p;

  if (debugtrace > 0)
     cout << "OnConnectionCleared " << token << endl;

  if (myConnection.line == NULL)
  {
      if (debugtrace > 1)
         cout << "line is NULL" << endl;
 //   return;
  }

  if (token != myConnection.gwcallToken)
  {
      if (debugtrace > 1)
         cout << "token not match" << endl;
	return;
  }

  p=myConnection.h323pvt;
  chan=myConnection.ast_chan;


  if (myConnection.line != NULL)
  {
      delete myConnection.line;
      myConnection.line = NULL;
  } 

	// check_h323_exception(...);
	if (chan != (struct ast_channel *)NULL)
	{
	    if (chan->state != AST_STATE_DOWN)
            {
                //cout << "OnConnection Cleared 0" << endl;
        	if (option_verbose > 1)
                    ast_log( LOG_WARNING, "Hungup 3 %s\n", chan->name );
                h323_hangup(chan);
		chan->pvt->pvt = NULL;
        	ast_softhangup(chan);
                chan->state = AST_STATE_DOWN;
		//if (chan->bridge)
		//  ast_softhangup(chan->bridge);
            }
            //chan->state = AST_STATE_DOWN;
	}

	//pthread_sleep(2);

        myConnection.h323pvt = (struct h323_pvt *)NULL;
        if (p != (struct h323_pvt *)NULL)
            free_h323_pvt(p);

}

GatewayConnection::GatewayConnection(GatewayEndPoint & _ep, 
                                     unsigned callReference)
  : H323Connection(_ep, callReference)
{
  SetMaxAudioDelayJitter(_ep.jitter);
  line = NULL;
  ast_chan = (struct ast_channel *)NULL;
  h323pvt = NULL;
  rx_chan = (H323Channel *)NULL;
  tx_chan = (H323Channel *)NULL;
}

GatewayConnection::~GatewayConnection()
{
    struct ast_channel *chan;
    struct h323_pvt *p = h323pvt;

    if (debugtrace > 0)
       cout << "destroy GatewayConnection " << gwcallToken << endl;

    if (line != NULL)
        delete line;

    line = NULL;

//    if (dfd != -1)
//       close(dfd);

    if (p != (struct h323_pvt *)NULL)
    {
  //  ClearCall(H323Connection::EndedByLocalUser);
      if (debugtrace > 0)
         cout << "free_h323_pvt" << endl;
    
       free_h323_pvt(p);
    }

    h323pvt = NULL;
    ast_chan = NULL;
    if (debugtrace > 0)
       cout << "DONE: destroy GatewayConnection" << endl;
}

void GatewayConnection::OnUserInputString(const PString & value)
{
  if (option_verbose > 3)
    cout << "OnUserInputString" << endl;
  if (line == NULL)
    return;

  OpalLineInterfaceDevice & device = line->GetDevice();
  unsigned lineNumber = line->GetLineNumber();

  if (option_verbose > 3)
    cout << "Line " << device << '-' << lineNumber << " user input: " << value <<
endl;
  device.PlayDTMF(lineNumber, value);
  // check_h323_exception(...);
}

void GatewayConnection::StartFastStartChannel(H323LogicalChannelList & fastStartChannels,
               unsigned sessionID, H323Channel::Directions direction)
{
  for (PINDEX i = 0; i < fastStartChannels.GetSize(); i++) 
  {
    H323Channel & channel = fastStartChannels[i];
    if (debugtrace > 1)
      cout << "channel no=" << channel.GetNumber() << " Session Id=" << channel.GetSessionID() << endl;
    if (channel.GetSessionID() == sessionID && channel.GetDirection() == direction) {
      if (debugtrace > 1)
        cout << "channel no=" << channel.GetNumber() << "Started" << endl;
      fastStartChannels[i].Start();
      break;
    }
  }
}


void GatewayConnection::SelectFastStartChannels(unsigned sessionID,
                                             BOOL transmitter,
                                             BOOL receiver)
{
  // Select all of the fast start channels to offer to the remote when initiating a call.


  for (PINDEX i = 0; i < (PINDEX)localCapabilities.GetSize()-1; i++) 
  {
    H323Capability & capability = localCapabilities[i];

    if (debugtrace > 1)
       cout << "2." << i << " capability=" << capability.GetFormatName() << endl;
    if (capability.GetDefaultSessionID() == (unsigned)sessionID) 
    {
      if (receiver) 
      {
        if (!OpenLogicalChannel(capability, sessionID,
		 H323Channel::IsReceiver)) 
	{
          cout << "3.1 SelectFastStartChannels, OpenLogicalChannel rx failed: " << capability << endl;
        }
      }

      if (transmitter) 
      {
        if (!OpenLogicalChannel(capability, sessionID, H323Channel::IsTransmitter)) 
	{
           cout << "4 SelectFastStartChannels, OpenLogicalChannel tx failed: " << capability << endl;
        }
      }
    }
  }
}



H323Capability *GatewayConnection::SelectLocalCapabilities(
		  H323Capability & cap) 
{
     H323Capability::CapabilityDirection cd =
                    cap.GetCapabilityDirection();

     if (debugtrace > 0)
     {
        cout << "SelectLocalCapabilities-" << cap.GetFormatName() << endl;
        cout << "localCapabilities Size=" << localCapabilities.GetSize() << endl;
     }

     for (PINDEX j=0; j < localCapabilities.GetSize(); j++)
     {
         if (debugtrace > 1)
            cout << "localCaps " << j << " =" << localCapabilities[j] << endl;

         if ( localCapabilities[j].GetMainType() == 
		H323Capability::e_Audio &&
                     localCapabilities[j] == cap )
         {
              if (debugtrace > 2)
                cout << "SelectLocalCapability 1 Ended=" << j << ": " << localCapabilities[j] << endl;
                return &localCapabilities[j];
         }
    }         

    return (H323Capability *)NULL;
}


void GatewayConnection::OnSelectLogicalChannels()
{
  if (debugtrace > 0)
     cout << "OnSelectLogicalChannels ID=" << gwcallToken << endl;

//   H323Connection & myConnection = (H323Connection &)*this;

  // Select the first codec that uses the "standard" audio session.
  switch (fastStartState) 
  {
     default : //FastStartDisabled :
      // No FastState
      if (debugtrace > 1)
         cout << "FastStartDisabled" << endl;
      if (rx_chan != NULL && tx_chan == NULL)
      {
         if (debugtrace > 1)
            cout << "FastStartDisabled 2" << endl;
         const H323Capability &cap = rx_chan->GetCapability();
         if (!OpenLogicalChannel(cap, rx_chan->GetSessionID(), 
			             H323Channel::IsTransmitter)) 
	   if (debugtrace > 1)
              cout << "OnSelectLogicalChannels, OpenLogicalChannel tx failed: " 
				<< cap << endl;
      }
      else
      if (rx_chan == NULL && tx_chan != NULL)
      {
         const H323Capability &cap = tx_chan->GetCapability();
	 if (debugtrace > 1)
            cout << "FastStartDisabled 3" << endl;
         if (!OpenLogicalChannel(cap, tx_chan->GetSessionID(), 
						H323Channel::IsReceiver)) 
	   if (debugtrace > 1)
               cout << "OnSelectLogicalChannels, OpenLogicalChannel rx failed: " << cap << endl;
      }
      else 
      {    // Note: the remoteCapabilities list is preference ordered
           int  selected[2] = {9999,0};
	   CodecXRefInfo *xf;
           H323Capability  *c,*chosen = NULL;
           for (PINDEX i=0; i < remoteCapabilities.GetSize(); i++)
           {
	       if ((c=SelectLocalCapabilities(remoteCapabilities[i])) != NULL )
               {
                   PString fmt = c->GetFormatName();
  		   PINDEX i = fmt.Find("{",0);
                   
                   xf = IsValidCodec( (const char *)fmt.Left(i) );
                   if (xf != NULL)
                   {
                     if (selected[0] > xf->preference)
                     { 
                           chosen = c;
			   selected[0] = xf->preference;
			   selected[1] = 0;
		     }
		   }
               }
           }

	   if (chosen != (H323Capability *)NULL)
	   {
	       if (debugtrace > 1)
	         cout << "Selected Caps=" << chosen->GetFormatName() << endl;
               OpenLogicalChannel(*chosen, 
                         RTP_Session::DefaultAudioSessionID, 
	                	H323Channel::IsTransmitter);
	       return;
	   }

	  ast_log(LOG_WARNING, "Failed localCapabilities local=%d remote=%d\n",
               localCapabilities.GetSize(),
			remoteCapabilities.GetSize() );

      }  // if .. else

     // SelectDefaultLogicalChannel(RTP_Session::DefaultAudioSessionID);
     // above cause crash

//      if (endpoint.CanAutoStartTransmitVideo())
//        SelectDefaultLogicalChannel(RTP_Session::DefaultVideoSessionID);
//      if (endpoint.CanAutoStartTransmitFax())
//        SelectDefaultLogicalChannel(RTP_Session::DefaultFaxSessionID);
       break;

    case FastStartInitiate :
      if (debugtrace > 1)
        cout << "FastStart Initiate" << endl;
      SelectFastStartChannels(RTP_Session::DefaultAudioSessionID, TRUE, TRUE);
      break;

    case FastStartResponse :
      if (debugtrace > 1)
         cout << "FastStart Response" << endl;

      StartFastStartChannel(fastStartChannels, 
	      RTP_Session::DefaultAudioSessionID, H323Channel::IsTransmitter);
      StartFastStartChannel(fastStartChannels, 
	     RTP_Session::DefaultAudioSessionID, H323Channel::IsReceiver);
      break;
  }

  if (debugtrace > 1)
    cout << "OnSelectLogicalChannels Exit 3" << endl;
}


BOOL GatewayConnection::OnStartLogicalChannel(H323Channel & channel)
{
  if (debugtrace > 0)
  {
     cout << "ID=" << gwcallToken << endl;
     cout << "Line " << line->GetDevice() << '-' << line->GetLineNumber()
       << " started logical channel: ";
  }

  PString fmt = channel.GetCapability().GetFormatName();
//  PINDEX i = fmt.Find("{",0);

 // char s[20];
  switch (channel.GetDirection()) {
    case H323Channel::IsTransmitter :
      if (debugtrace > 0)
          cout << "sending nativeformats=" << h323pvt->owner->nativeformats << endl;;
      tx_chan = &channel; 

      {
	line->SetReadFormat(0,OPAL_PCM16);
        ast_chan->nativeformats = AST_FORMAT_SLINEAR;
	h323pvt->lastoutput = AST_FORMAT_SLINEAR;
        struct CodecXRefInfo *cf = IsValidCodec((const char *)fmt);
        if (debugtrace > 1)
             cout << "conn " << ast_chan->name << " format=" << ast_chan->nativeformats << endl;
        if (cf != NULL)
        {
            line->out_min = cf->out_min;
	    line->out_max = cf->out_max;
	    h323pvt->tx_framerate = cf->rxframes;
        }
      }
      break;

    case H323Channel::IsReceiver :
      if (debugtrace > 0)
          cout << "receiving nativeformats=" << (h323pvt->owner)->nativeformats << endl; 
      rx_chan = &channel; 
      
      {
	line->SetWriteFormat(0,OPAL_PCM16);
        (h323pvt->owner)->nativeformats = AST_FORMAT_SLINEAR;
	h323pvt->lastinput = AST_FORMAT_SLINEAR;
        struct CodecXRefInfo *cf = IsValidCodec((const char *)fmt);
        if (cf != NULL)
        {
            line->in_min = cf->in_min;
	    line->in_max = cf->in_max;
	    h323pvt->rx_framerate = cf->txframes;
	}
      }

      break;

    default :
      break;
  }

  if (debugtrace > 0)
     cout << channel.GetCapability() << endl;
 
  if (ast_chan->state < AST_STATE_UP && 
		(rx_chan != NULL && tx_chan != NULL))
  {
     ast_chan->state = AST_STATE_UP;
     if (debugtrace > 1)
        cout << "WE ARE AST_STATE_UP (" << gwcallToken << ")" << endl;

     write(h323pvt->infd[1],"ASTUP\n\0",6);
     ioctl(h323pvt->infd[1],FDFLUSH);
     pthread_sleep(1);
  }

  return TRUE;
}


void GatewayConnection::OnCleared()
{
    if (debugtrace > 0)
    {
       cout << "GatewayConnection::OnCleared " << GetCallToken() << endl;

      switch( GetCallEndReason() )
      {
      case EndedByLocalUser:    
	   cout << "Local endpoint application cleared call" << endl;
	   break;
      case EndedByNoAccept:     
	   cout << "Local endpoint did not accept call FALSE" << endl;
	   break;
      case EndedByAnswerDenied:      
	   cout << "Local endpoint declined to answer call" << endl;
	   break;
      case EndedByRemoteUser:        
	   cout << "Remote endpoint application cleared call" << endl;
	   break;
      case EndedByRefusal:           
	   cout << "Remote endpoint refused call" << endl;
	   break;
      case EndedByNoAnswer:      
	   cout << "Remote endpoint did not answer in required time" << endl;
	   break;
      case EndedByCallerAbort:       
	   cout << "Remote endpoint stopped calling" << endl;
 	   break;
      case EndedByTransportFail:    
	   cout << "Transport error cleared call" << endl;
	   break;
      case EndedByConnectFail:       
	   cout << "Transport connection failed to establish call" << endl;
	   break;
      case EndedByGatekeeper:        
	   cout << "Gatekeeper has cleared call" << endl;
	   break;
      case EndedByNoUser:            
	   cout << "Call failed as could not find user (in GK)" << endl;
	   break;
      case EndedByNoBandwidth:          
	   cout << "Call failed as could not get enough bandwidth" << endl;
	   break;
      case EndedByCapabilityExchange:   
	   cout << "Could not find common capabilities" << endl;
	   break;
      case EndedByCallForwarded:    
	   cout << "Call was forwarded using FACILITY message" << endl;
	   break;
      case EndedBySecurityDenial:    
	   cout << "Call failed a security check and was ended" << endl;
	   break;
      case EndedByLocalBusy:         
	   cout << "Local endpoint busy" << endl;
	   break;
      case EndedByLocalCongestion:   
	   cout << "Local endpoint congested" << endl;
	   break;
      case EndedByRemoteBusy:        
	   cout << "Remote endpoint busy" << endl;
	   break;
      case EndedByRemoteCongestion:  
	   cout << "Remote endpoint congested" << endl;
	   break;
      case EndedByUnreachable:       
	   cout << "Could not reach the remote party" << endl;
	   break;
      case EndedByNoEndPoint:        
	   cout << "The remote party is not running an endpoint" << endl;
	   break;
      case EndedByHostOffline:       
           cout << " The remote party host off line" << endl;
	   break;
      case NumCallEndReasons:
      default: 
	   cout << "No Reason Given" << endl;
	   break;
      }
    }
    h323ep->OnConnectionCleared(*this, GetCallToken());   //gwcallToken);
   
}


void GatewayConnection::OnLogicalChannelJitter(
      H323Channel * channel,   /// Channel that is to be limited
      DWORD jitter,            /// Estimated received jitter in microseconds
      int skippedFrameCount,   /// Frames skipped by decodec
      int additionalBuffer     /// Additional size of video decoder buffer
    )
{
    PString fmt = channel->GetCapability().GetFormatName();
    if (debugtrace > 0)
       cout << "OnLogicalChannelJitter " << fmt << " Jitter Req=" << jitter << endl;
}

PNXGw::PNXGw()
  : PProcess("File-Sure VOIP Project", "PNXGw",
             MAJOR_VERSION, MINOR_VERSION, BUILD_TYPE, BUILD_NUMBER)
{
   h323ep = NULL;
   srand(time(NULL));
}

PNXGw::~PNXGw()
{
  delete h323ep;
}

void PNXGw::Main()
{
}

#include "chan_h323_pnx.cxx"

